"""Main module for the site class"""
#! /usr/bin/env python

# Python imports
import sys
import os
import base64

# Module imports
import numpy as np
from dash import Dash, dcc, html, Input, Output
import plotly.express as px
import pandas as pd

# Local imports
BASE_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(BASE_PATH)
import visnet1d

class Site:
    """Site class that handles the dash web app"""

    def __init__(self, function, boundaries, input_dict_list=None, **kwargs):
        """ Initialises a site to visualise a vasculature network.

        Args:
            function (object) : A function that takes a set of kwarg arguments and produces an
                output.
            boundaries (dict) : A dictionary of each location's start and end.
                Expected as a dictionary with:
                    "location name" : [start, end]
                for each location / edge.
            input_dict_list (list, optional) : A list of dictionaries where each dictionary has the fields:
                name (str) : Name of variable to be displayed to site.
                id (str) : Name to be passed to the function (also used as ID).
                value (object) : Default value to be displayed to site.
                type (str) : Type of the value.
                If None, doesn't display any inputs. Defaults to None.
        kwargs:
            plot (function, optional) : The function to plot the graph.
                Defaults to visnet1d.site._plot_x().
            xlab (str, optional) : The label for the x-axis. Defaults to "Value".
            ylab (str, optional) : The label for the y-ayis. Defaults to "Value".
            title (str, optional) : The title of the plot. Defaults to "signal".
            gwidth (int, optional) : The percentage of screen width occupied by the graph.
                Defaults to 70.
            image (str, optional) : Path to the image to load under the inputs section.
                If None, no image will be displayed. Defaults to None.
            iheight (int, optional) : The percentage of the screen height occupied by the image.
                Defaults to 80.
        """

        # Sets pandas plotting backend
        pd.options.plotting.backend = "plotly"

        # Loads keyword arguments
        self.plot = kwargs.get('plot', None)
        self.xlab = kwargs.get('xlab', "Value")
        self.ylab = kwargs.get('ylab', "Time (s)")
        self.title = kwargs.get('title', "Signal")

        gwidth = kwargs.get('gwidth', 70)
        image = kwargs.get('image', None)
        iheight = kwargs.get('iheight', 80)

        # Defines static methods/attributes
        self.app = Dash(__name__)
        self.input_dict_list = input_dict_list
        self.function = function
        self.boundaries = boundaries

        # Sets up a dummpy input which will be hidden.
        if input_dict_list is None:
            input_dict_list = [{"name" : "", "id": "foo", "value" : '', "type" : "hidden"}]

        # Defines dynamic methods/attributes
        self.function_kwargs = get_kwargs(input_dict_list)
        self.loc_slider_vals = dict()
        self.loc_percentages = dict()
        self.loc_keys = []

        locs = list(boundaries.keys())
        loc_dd_id = "loc-dropdown"
        loc_radio_id = "radio-locs"
        selected_locs = [locs[0]]

        self.app.layout = html.Div([
            html.Div([
                html.H1("Inputs"),
                html.Div([
                    "Location: ",
                    dcc.Dropdown(locs, locs[0], id=loc_dd_id, multi=True),
                ]),
                visnet1d.layout.inputs_from_dict_list(
                    input_dict_list, style={'padding':10, 'flex':1},
                ),
                visnet1d.layout.format_image(image, height=iheight),
            ], style={'padding':10, 'width': f"{100 - gwidth}vw"}, draggable=True),
            html.Div([
                html.H1("VisNet1D", style = {'textAlign': 'center'}),
                html.Div([
                    html.Label('X position'),
                    dcc.Slider(
                        min=0,
                        max=1,
                        value=0,
                        vertical=False,
                        id="slider-x",
                    ),
                    html.Div(id="slider-x-label"),
                    dcc.Dropdown(
                        selected_locs, 
                        selected_locs[0], 
                        id=loc_radio_id, 
                    ),
                ]),
                html.Div([
                    dcc.Graph(
                        id='graph-x', 
                        config={"editable" : True, "modeBarButtonsToAdd" : ["hoverCompareCartesian"]},
                    ),
                ]),
            ], style={'padding':10, 'width': f"{gwidth}vw"}, draggable=True),
        ], style={'display': 'flex', 'flex-direction': 'col'})

        inputs = [Input(item["id"], 'value') for item in input_dict_list]
        input_kwarg_keys = [item["id"] for item in input_dict_list]

        ### Callback functions ###
        @self.app.callback(
            Output('slider-x', 'min'),
            Output('slider-x', 'max'),
            Output('slider-x', 'value'),
            Input(loc_radio_id, 'value'),
        )
        def _set_slider_range(loc_key):
            """Gets the slider range along with the selected slider value"""
            v_start, v_end = get_relative_x(*boundaries[loc_key])
            try:
                v_selected = self.loc_slider_vals[loc_key]
            except KeyError:
                v_selected = (v_end - v_start) / 2
            return v_start, v_end, v_selected

        @self.app.callback(
            Output('slider-x-label', 'children'),
            Input('slider-x', 'value'),
            Input(loc_radio_id, 'value'),
            Input(loc_dd_id, 'value'),
        )
        def _update_slider(x_pos, selected_loc_key, loc_keys):
            """Display the percentage along the location"""
            v_start, v_end = get_relative_x(*boundaries[selected_loc_key])
            self.loc_percentages[selected_loc_key] = (x_pos - v_start) / (v_end - v_start) * 100
            if not isinstance(loc_keys, list):
                loc_keys = [loc_keys]
            rtn_str = [f"{key} = {self.loc_percentages[key]:.0f}%" for key in loc_keys]
            return ", ".join(rtn_str)

        @self.app.callback(
            Output(loc_radio_id, "options"),
            Output(loc_radio_id, "value"),
            Input(loc_dd_id, "value"),
        )
        def _update_radio_options(loc_options):
            """Updates the list of selected locations"""
            if not isinstance(loc_options, list):
                loc_options = [loc_options]
            return loc_options, loc_options[-1]

        @self.app.callback(
            Output('graph-x', 'figure'),
            Input('slider-x', 'value'),
            Input(loc_radio_id, 'value'),
            Input(loc_dd_id, 'value'),
            *inputs,
        )
        def _get_figure_x(x, selected_loc_key, loc_keys, *input_kwarg_values): # pylint: disable=invalid-name
            """Plots the figure with up to date values"""

            # Updates internal variables
            self.update_kwargs(input_kwarg_keys, input_kwarg_values)
            self.x = x
            self.loc_slider_vals[selected_loc_key] = x
            if isinstance(loc_keys, list):
                self.loc_keys = loc_keys
            else:
                self.loc_keys = [loc_keys]

            if self.plot is None:
                return self._plot()
            return self.plot(self)

    def update_kwargs(self, keys, values):
        """Updates the function kwargs dictionary

        Also updates internal property `self.function_kwargs`.

        Args:
            keys (list) : A list of function key word arguments.
            values (list) : A list of associated values.
                Expected to be in the same order as keys.

        Returns:
            function_kwargs (dict) : A dictionary of kwarg pairs.
        """

        function_kwargs = {}
        for i, key in enumerate(keys):
            function_kwargs[key] = values[i]

        self.function_kwargs = function_kwargs

        return function_kwargs

    def run(self, **kwargs):
        """Runs the server

        Key word arguments are directly passed to self.app.run_server()
        """
        return self.app.run_server(**kwargs)

    def _plot(self):
        """ Plots the graph

        Returns:
            fig (plotly figure) : Figure to displayed to the site.
        """

        data = dict()
        for i, key in enumerate(self.loc_keys):
            # pylint: disable=invalid-name
            y, x = self.function(self.loc_slider_vals[key], **self.function_kwargs)
            if i == 0:
                data[self.xlab] = x
            data[key] = y
        df = pd.DataFrame(data)
        fig = df.plot(
                x=self.xlab, 
                y=self.loc_keys, 
                title=self.title, 
                kind='line',
                labels=dict(value=self.ylab, variable="Location"),
                )

        return fig

def get_kwargs(input_dict_list):
    """Gets a dict of kwarg from a list of dictionaries

    Args:
        input_dict_list (list) : A list of dictionaries with the format "id" : "value".

    Returns:
        kwarg_dict (dict) : A dictionary of kwarg pairs.
    """

    kwarg_dict = {}
    for item in input_dict_list:
        key, value = item["id"], item["value"]
        kwarg_dict[key] = value

    return kwarg_dict

def get_relative_x(start, end):
    """Gets the relative distance between a start and an end"""
    return 0, end - start

def get_static_function(output_mat, t, axis=0):
    """Used for static inputs

    Args:
        output_mat (numpy array) : This is the output of some pre-run function.
            The use cases would be explore this output at various locations with
            the need to continually evaluate a function.
        t (numpy array) : The time axis to return which is used for plotting.
        axis (int, optional) : If axis is 0, returns output_mat[x, :].
            Else returns output_mat[:, x].
            Defaults to 0.

    """
    def fn(x, **kwargs):
        if axis == 0:
            return output_mat[int(x), :], t
        return output_mat[:, int(x)], t
    return fn



#! /usr/bin/env python

# Module imports
from dash import Dash, dcc, html, Input, Output
import plotly.express as px
from PIL import Image

def inputs_from_dict_list(input_list, style=None): 
    """Gets an input layout from a list of dictionary

    Args: 
        input_list (list) : A list of dictionaries where each dictionary has the fields:
            name (str) : Name of variable to be displayed to site.
            id (str) : Name to be passed to the function (also used as ID).
            value (object) : Default value to be displayed to site.
            type (str) : Type of the value.
        style (dict) : The default style for the dash.html.Div object.
            Default to None.
    Returns:
        input_layout (Dash html object) : A Dash html div object.
    """

    input_fields = []
    for dict_item in input_list:
        input_name = f"{dict_item['name']}: " if len(dict_item['name']) > 0 else ""
        field = html.Div([
            input_name,
            dcc.Input(id=dict_item["id"], value=dict_item["value"], type=dict_item["type"]),
        ])
        input_fields.append(field)

    return html.Div(input_fields, style=style)

def format_image(path, height=80):
    if path is None:
        return html.Div()
    config = {
        "modeBarButtonsToAdd": [
            "drawline",
            "drawopenpath",
            "drawclosedpath",
            "drawcircle",
            "drawrect",
            "eraseshape",
            "orbitRotation",
        ],
    }
    fig = px.imshow(Image.open(path))
    fig.update_xaxes(showticklabels=False)
    fig.update_yaxes(showticklabels=False)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=0))
    return dcc.Graph(figure=fig, config=config, style={'height': f"{height}vh"})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['visnet1d']

package_data = \
{'': ['*']}

install_requires = \
['dash>=2.3.1,<3.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pillow>=9.4.0,<10.0.0',
 'plotly>=5.7.0,<6.0.0']

setup_kwargs = {
    'name': 'visnet1d',
    'version': '0.1.4',
    'description': 'A visualation tool for 1D networks',
    'long_description': '# VisNet1D\n\nA simple visualisation tool for 1D networks. \n\n![An image of visnet1d](visnet1d.png)\n\n## Installation\n\n### Use in your project\n\nPython Package Index (PIP):\n```\npip install visnet1d\n```\n\nPoetry:\n```\npoetry add visnet1d\n```\n\n### Use the development version\n\nGit:\n```\ngit clone git@gitlab.com:abdrysdale/visnet1d.git\nnix-shell\npoetry install\npoetry shell\n```\n\nInstallation instructions for `nix` can be obtained from [here](https://nixos.org/download.html).\nIf you don\'t have and don\'t want to install nix, it is sufficient to just install poetry and skip the `nix-shell` command.\nPoetry installation instructions can be obtained from [here](https://python-poetry.org/).\n\n### Just want to run the examples?\n\nIf you just want to run one of the examples, such as the vessel network example:\n```\npip install visnet1d\ngit clone git@gitlab.com:abdrysdale/visnet1d.git\ncd visnet1d\n```\n\nThe run the desired example script with:\n```\npython examples/vessel_network.py\n```\nNaturally replacing `vessel_network.py` with any of the other examples.\nNext, go to `http://127.0.0.1:8050` in your browser and start exploring!\n\n\n## Usage\n\nVisNet1D is intended to be used to visualise signals across a 1-dimensional network at potentially multiple locations.\nIt can either be used to explore an output of an other function, viewing signal(s) at different locations and comparing the results or it can be used to visualise the change in a function output due to a change in inputs.\n\nAn example usage is shown below:\n\nFirst import the library:\n```python\nimport visnet1d\n```\n\n\nNow we define a simple function that takes a spatial argument along with other keyword arguments.\nThis function returns y as a function of t at some spatial point x.\nIt must return two vectors to be compatible with the default plotting function.\nBut don\'t worry! You can define your own plotting functions - we\'ll get to that later.\n```python\ndef foobar(x, tmax=1, a=2, b=3):\n    t = np.linspace(0, tmax, 100)\n    y = a * np.sin(x + 2 * np.pi * t / b)\n    return y, t\n```\n\nHere we define a list for the function input arguments.\nThese are used to display the inputs on the web-page and pass the results to the function `foobar`.\n```python\ninput_dict_list = [\n    {\n        "name" : "Amplitude",   # Name to be displayed on the web app.\n        "id" : "a",             # Internal id and keyword argument for the function (foobar).\n        "value" : 1,            # Initial value.\n        "type" : "number",      # Used by Dash to determine the input type.\n    },\n    {\n        "name" : "Period",\n        "id" : "b",\n        "value" : 2,\n        "type" : "number",\n    },\n    {\n        "name" : "Maximum t",\n        "id" : "tmax",\n        "value": 1,\n        "type" : "number",\n    },\n]\n```\nThe allowed types that dash input accepts can be found [here](https://dash.plotly.com/dash-core-components/input).\n\nNext we need to define the boundaries that separate locations.\nTo illustrate this, let\'s consider two vessels connected in series:\n```python\nboundaries = {     # Locations can be selected from a drop-down menu\n        "Vessel 1" : [0, np.pi],          # [Start, End]\n        "Vessel 2" : [np.pi, 3 * np.pi],  \n}\n```\n\nNow we create a site object and run the site.\n```python\nsite = visnet1d.Site(\n        foobar,             # Function to plot.\n        boundaries,         # Locations to select.\n        input_dict_list,    # List of function inputs (each input has it\'s own dictionary).\n        xlab="Time (s)",    # X-axis label\n        ylab="Response",    # Y-axis label\n        title="My title",   # Title\n)\n\n# Runs the server with the default arguments - same as site.run()\nsite.run(host="127.0.0.1", port="8050", debug=False)\n```\nGo to `127.0.0.1:8050` on your web browser view the webapp.\n\n### Custom plots\n\nCustom plot functions can be utilised using the `plot` keyword argument for the visnet1d.Site class.\nE.g.\n\n```python\nsite = visnet.Site(\n    my_function,\n    my_boundaries,\n    my_input_dictionary_list,\n    plot=my_custom_plot_function, # <- This is how you pass your custom plot function.\n)\n```\n\nA few things to note about writing custom plotting functions:\n\n- The function must take the Site object as the **only** input argument.\n- The function must return a plotly figure.\n\nA few notable attributes of the Site class:\n\n- `site.function_kwargs` : Is a list of dictionaries of the inputs defined by the `input_dict_list` argument.\n    These are the values that you can change in the inputs section in the webapp.\n- `site.boundaries` : The boundaries defined by the `boundaries` argument.\n- `site.loc_slider_vals` : A dictionary of all of the slider positions for all of the previously selected locations.\n    The dictionary is of the format `location` : `value`. Where location is the location key used in the `boundaries` argument and `value` is the value of the slider.\n- `site.loc_keys` : A list of the locations currently selected. Where each location is the key used in the `boundaries argument.`\n- `site.x` : The current position of the active slider.\n\n### Exploring a static output\n\nBy default, VisNet1D evaluates the function each time an input is changed.\nFor a very expensive function this isn\'t ideal.\nMoreover, there might be use case whereby the outputs of very complicated network have been produced by some other expensive function and you just wish to use VisNet1D to explore the output.\n\nHere\'s how you\'d do something like that:\n```python\nmy_outputs = visnet1d.get_static_function(out_mat, t, axis=0)\n```\nFirst use the inbuilt function `get_static_function()` to handle static outputs.\nThis function needs to be passed the 2D output matrix and the value for the x-axis.\nThe output matrix might be something like a signal as a function of space and time.\n\nMoreover, as the outputs are static there are no function inputs so we can do away with the `input_dict_list` argument.\nThis is done like so:\n```python\nsite = visnet.Site(\n    my_outputs,\n    my_boundaries,\n    input_dict_list=None,\n)\n```\n\n### Customising plots\n\nBy default all the plots are editable and you can pass the x axis label, y axis label and title into the site object with the following:\n```python\nsite = visnet.Site(\n    my_function,\n    my_boundaries,\n    my_input_dictionary_list,\n    xlab=my_x_label,\n    ylab=my_y_label,\n    title=my_title,\n)\n```\n\nMoreover, the width of the graph itself can be controlled by using:\n```python\nsite = visnet.Site(\n    my_function,\n    my_boundaries,\n    my_input_dictionary_list,\n    gwidth=80, # This sets the width of the graph to be 80% of the screen width.\n)\n\n```\n\nIt is also possible to display an optional image underneath the input section to aid network visualisation:\n```python\nsite = visnet.Site(\n    my_function,\n    my_boundaries,\n    my_input_dictionary_list,\n    image=\'path/to/my/image.png\'    # Any Pillow supported image format can be used\n    iheight=80,                     # Adjusts the image height to be 80% of the screen height.\n)\n```\n\n## Examples\n\nFor full examples see the [examples/](examples/) directory.\n\n## Contributing\n\n- For contributing guidelines see [CONTRIBUTING](CONTRIBUTING.md).\n',
    'author': 'Alex',
    'author_email': 'adrysdale@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/abdrysdale/visnet1d',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
